/* jshint esversion: 8 */

const path = require('path');
const yaml = require('yaml');
const _ = require('lodash');
const { v4: uuid } = require('uuid');

function buildVegaNoZoomFitCode(eId) {
	let code = '';
	code += `var vegaViewBox = document.getElementById('vega-view-${eId}');`;
	code += `var contentWrapper = vegaViewBox.closest('.slot-content-wrapper');`;
	code += `contentWrapper.style.overflow = 'auto';`;
	return code;
}

function buildVegaCode(spec, vegaOptions) {
	const eId = uuid();
	const padding = vegaOptions.padding;
	const paddingProp = typeof padding === 'number' || typeof padding === 'object'
		? ' padding: ' + JSON.stringify(padding) + ','
		: ''
	if (typeof vegaOptions.width === 'number') spec.width = vegaOptions.width;
	if (typeof vegaOptions.height === 'number') spec.height = vegaOptions.height;
	const zoomFitCode = buildVegaNoZoomFitCode(eId);
	const renderer = vegaOptions.renderer || 'svg';
	const hover = vegaOptions.hover === undefined ? true : vegaOptions.hover;
	const result = `<div id="vega-view-${eId}" class="vega-view">
	Loading Vega Visualization...
</div><script>boomack.run(function () {
	var spec = ${JSON.stringify(spec)};
	spec = boomack.replaceThemeProperties(spec);
	var vegaView = document.getElementById('vega-view-${eId}');
	var wrapper = $(vegaView.closest('.slot-content-wrapper'));
	if (!spec.width) spec.width = wrapper.innerWidth() -
		parseFloat(wrapper.css('padding-left')) -
		parseFloat(wrapper.css('padding-right'));
	if (!spec.height) spec.height = wrapper.innerHeight() -
		parseFloat(wrapper.css('padding-top')) -
		parseFloat(wrapper.css('padding-bottom')) - 8;
	try {
		vegaEmbed('#vega-view-${eId}', spec, {
			config: boomack.vegaConfig,
			renderer: '${renderer}',${paddingProp}
			actions: false, hover: ${hover},
			tooltip: { theme: 'boomack' },
		}).then(function ({view, spec, vgSpec, finalize}) {
			${zoomFitCode}
		}).catch(function (err) {
			vegaView.outerHTML = '<div class="ui error message"><div class="header">Error in Vega specification</div><div class="content">' + err.message + '</div></div>';
		});
});</script>`;
	return result;
}

function textTransformationVega (text, {options}) {
	const spec = yaml.parse(text);
	const vegaOptions = options.vega || {};
	return {
		text: buildVegaCode(spec, vegaOptions),
		requires: ['vega', 'vega-embed'],
	};
}

function rendererVega(url, {options}) {
	const vegaOptions = _.omit(_.get(options, 'vega') || {}, ['spec']);
	const spec = _.get(options, ['vega', 'spec']);
	const dataTemplates = _.filter(_.get(spec, ['data']), d => _.get(d, 'url') === null);
	for (const dt of dataTemplates) {
		dt.url = url;
	}
	return {
		html: buildVegaCode(spec, vegaOptions),
		requires: ['vega', 'vega-embed'],
	};
}

const defaultDocumentHeight = 384;

function buildVegaLiteZoomFitCode(eId) {
	return `
	boomack.registerSlotResizeHandler('#vega-view-${eId}', function (size, zoomedSize) {
		var vegaViewBox = document.getElementById('vega-view-${eId}');
		var contentBox = vegaViewBox.closest('.slot-content');
		var contentWrapper = contentBox.closest('.slot-content-wrapper');
		contentBox.style.overflow = 'hidden';
		contentWrapper.style.overflow = 'hidden';
		var width = Math.floor(zoomedSize.width);
		var height = null;
		if (isDocumentLayout) {
			height = Math.floor(zoomedSize.width / size.width * spec.height);
			contentBox.style.height = height + 'px';
			contentWrapper.style.height = (spec.height + paddingY) + 'px';
		} else {
			height = Math.floor(zoomedSize.height);
		}
		view._width = view._viewWidth = width;
		vegaViewBox.style.width =  width + 'px';
		view._height = view._viewHeight = height;
		vegaViewBox.style.height =  height + 'px';
		// trigger resizing of Vega-Lite
		window.dispatchEvent(new Event('resize'));
	}, true);`;
}

function buildVegaLiteNoZoomFitCode(eId, scale, align) {
	let code = '';
	code += `var vegaViewBox = document.getElementById('vega-view-${eId}');`;
	code += `var contentBox = vegaViewBox.closest('.slot-content');`;
	code += `var contentWrapper = contentBox.closest('.slot-content-wrapper');`;
	if (!scale) {
		code += `contentBox.style.width = '100%';`;
		code += `contentBox.style.height = '100%';`;
	}
	code += `contentWrapper.style.overflow = 'auto';`;
	// trigger resizing of Vega-Lite
	code += `window.dispatchEvent(new Event('resize'));`;
	return code;
}

function buildVegaLiteCode(spec, vegaOptions, options) {
	const eId = uuid();
	const padding = vegaOptions.padding;
	const paddingProp = typeof padding === 'number' || typeof padding === 'object'
		? ' padding: ' + JSON.stringify(padding) + ','
		: ''
	if (typeof vegaOptions.width === 'number') spec.width = vegaOptions.width;
	if (typeof vegaOptions.height === 'number') spec.height = vegaOptions.height;
	const zoomFit = (!options.scale || options.scale === 'none')
		&& vegaOptions.zoomFit !== false
		&& (spec.width === 'container' || spec.height === 'container');
	const zoomFitCode = zoomFit
		? buildVegaLiteZoomFitCode(eId)
		: buildVegaLiteNoZoomFitCode(eId, options.scale, options.align);
	const renderer = vegaOptions.renderer || 'svg';
	const hover = vegaOptions.hover === undefined ? true : !!vegaOptions.hover;
	const result = `<div id="vega-view-${eId}" class="vega-view">
	Loading Vega Lite Visualization...
</div><script>
	boomack.run(function () {
		var isDocumentLayout = document.getElementById('panel').querySelector('.layout').classList.contains('layout-document');
		var vegaViewBox = document.getElementById('vega-view-${eId}');
		var contentWrapper = vegaViewBox.closest('.slot-content-wrapper');
		var wrapperStyle = window.getComputedStyle(contentWrapper, null);
		var paddingY =
			Number.parseFloat(wrapperStyle.getPropertyValue('padding-left')) +
			Number.parseFloat(wrapperStyle.getPropertyValue('padding-right'));
		var liteSpec = ${JSON.stringify(spec)};
		liteSpec = boomack.replaceThemeProperties(liteSpec);
		if (isDocumentLayout && (!liteSpec.height || liteSpec.height === 'container')) {
			var height = ${defaultDocumentHeight} + paddingY;
			var slotBox = vegaViewBox.closest('.slot-content-box');
			var slotBoxStyle = window.getComputedStyle(slotBox, null);
			var minHeight = Number.parseFloat(slotBoxStyle.getPropertyValue('min-height'));
			var exactHeight = slotBox.style.height ?
				Number.parseFloat(slotBoxStyle.getPropertyValue('height')) :
				Number.NaN;
			var maxHeight = Number.parseFloat(wrapperStyle.getPropertyValue('max-height'));
			if (Number.isFinite(exactHeight)) height = exactHeight;
			if (Number.isFinite(minHeight)) height = Math.max(height, minHeight);
			if (Number.isFinite(maxHeight)) height = Math.min(height, maxHeight);
			liteSpec.height = height - paddingY;
		}
		vegaEmbed('#vega-view-${eId}', liteSpec, {
			config: boomack.vegaConfig,
			renderer: '${renderer}',${paddingProp}
			actions: false, hover: ${hover},
			tooltip: { theme: 'boomack' },
		}).then(function ({view, spec, vgSpec, finalize}) {
			${zoomFitCode}
		}).catch(function (err) {
			vegaViewBox.outerHTML = '<div class="ui error message"><div class="header">Error in Vega-Lite specification</div><div class="content">' + err.message + '</div></div>';
		});
	});
</script>`;
	return result;
}

function textTransformationVegaLite (text, {options}) {
	const spec = yaml.parse(text);
	if (typeof spec.width === 'undefined') spec.width = 'container';
	if (typeof spec.height === 'undefined') spec.height = 'container';
	const vegaOptions = options.vega || {};
	return {
		text: buildVegaLiteCode(spec, vegaOptions, options),
		requires: ['vega', 'vega-lite', 'vega-embed'],
	};
}

function rendererVegaLite(url, {options}) {
	const vegaOptions = _.omit(_.get(options, 'vega') || {}, ['spec']);
	const spec = _.get(options, ['vega', 'spec']);
	_.set(spec, ['data', 'url'], url);
	return {
		html: buildVegaLiteCode(spec, vegaOptions, options),
		requires: ['vega', 'vega-lite', 'vega-embed'],
	};
}

module.exports = {
	id: 'vega',
	textTransformations: {
		'vega': textTransformationVega,
		'vega-lite': textTransformationVegaLite,
	},
	renderers: {
		'vega': rendererVega,
		'vega-lite': rendererVegaLite,
	},
	scripts: {
		'vega': [
			path.join(__dirname, 'vendor', 'vega.min.js'),
			path.join(__dirname, 'vega-config.js'),
		],
		'vega-lite': [
			path.join(__dirname, 'vendor', 'vega-lite.min.js'),
		],
		'vega-embed': [
			path.join(__dirname, 'vendor', 'vega-embed.min.js'),
		],
	},
	styles: {
		'vega': [
			path.join(__dirname, 'vega-overrides.css'),
		],
	},
	init: {
		types: {
			'application/x-vega-spec': {
				text: true,
				options: {
					transformation: 'vega',
				},
			},
			'application/x-vega-lite-spec': {
				text: true,
				options: {
					transformation: 'vega-lite',
				},
			},
		},
	},
};
